import json
from .logmessage import LogMessage 

class WebsocketMessage:
    def __init__(self):
        self.subject = None
        self.sender = None
        self.topics = []
        self.payload = None
        self.persist = False
    
    @staticmethod
    def fromRecv(message_text):
        try:
            data = json.loads(message_text) 
            obj = WebsocketMessage()
            obj.subject = data.get('subject')
            obj.sender = data.get('sender')
            obj.topics = data.get('topics')
            obj.payload = data.get('payload')
            obj.persist = data.get('persist')
            return obj
        except Exception as e:
            LogMessage(f'[ERROR] Error in WebsocketMessage =>  {e}')
            return WebsocketMessage()
    
    def toDict(self):
        return {
            "subject" : self.subject,
            "sender" : self.sender,
            "topic" : self.topics,
            "payload" : self.payload,
            "persist" : self.persist,
        }


from setuptools import setup

VERSION = '1.0.2' 
DESCRIPTION = 'Websocket Server for python'
LONG_DESCRIPTION = '''
@section{Python Package for Websocket Server Live}

1. WebsocketServer built in python.

2. For active server on command line:

- Install package with PIP;
- Run command wsserver-malbizer '<port>'
- Send JSON string for serve on model:

    {
        "subject": "subject for your message",
        "sender": "id_sender",
        "topic": [], //List of topics to inscribe message
        "payload": "Value for your message (JSON string)",
        "persist": false //indicate if message is persisted and send for new clients
    }



3. For using on python file:

- Import wsservermalbizer with WebsocketServer and use the code:

    WebSocketServer(port=port).start()
'''


# Setting up
setup(
       # 'name' deve corresponder ao nome da pasta 'verysimplemodule'
        name="wsserver_malbizer", 
        version=VERSION,
        author="Anderson Souza",
        author_email="anderson@malbizer.com.br",
        description=DESCRIPTION,
        long_description_content_type="text/markdown",
        long_description=LONG_DESCRIPTION,
        packages=['wsservermalbizer'],
        scripts=['bin/wsserver-malbizer.py'],
        zip_safe=False,
        install_requires=["websockets==9.1"], # adicione outros pacotes que 
        # precisem ser instalados com o seu pacote. Ex: 'caer'
        
        keywords=['python', 'websocket', 'server'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ]
)


import asyncio
from os import environ
import websockets
import json

from .logmessage import LogMessage
from .websocketmessage import WebsocketMessage 

class ArgsWebsocket:
    def __init__(self):
        self.__values__ = {}
        self.hashcode = None
        self.topics = None    
    
    @staticmethod
    def fromPath(path):
        try:
            args = ArgsWebsocket()
            args.__values__ = {p.split("=")[0]:p.split("=")[1] for p in path.split("?")[-1].split("&")}
            args.hashcode = args.__values__.get("hashcode")
            args.topics = args.__values__.get("topics").split(",")
            return args
        except:
            LogMessage("[ERROR] Invalid Args => Generate: ArgsWebsocket class")
            return ArgsWebsocket()

class WebSocketServer:
    def __init__(self, uri="127.0.0.1", port=int(environ.get("PORT") or "5589")):
        self.__users__ = set()
        self.__uri__ = uri
        self.__port__ = port
        self.__last_messages__ = {}
    
    async def register(self, websocket):
        self.__users__.add(websocket)
        try:
            for t in websocket.values.topics:
                if t in self.__last_messages__: await websocket.send(self.__last_messages__.get(t))
        except Exception as e:
            LogMessage(f'[ERROR] Register Error => {e}')
        
    async def unregister(self, websocket):
        try:
            self.__users__.remove(websocket)
            LogMessage(f"[INFO] {websocket.values.hashcode} close connection!")
        except Exception as e:
            LogMessage(f"[ERROR] Error on remove user! => {self.__class__.__name__} -> {str(e)} ")
            
    async def send_broadcast_message(self, message:WebsocketMessage):
        try:
            if self.__users__:  
                data = json.dumps(message.toDict(), indent=4,)
                await asyncio.wait([user.send(data) for user in self.__users__])
        except Exception as e:
            LogMessage(f'[ERROR] Error on broadcast message => {e}')

    async def send_private_message(self, message:WebsocketMessage):
        try:
            if self.__users__:  
                LogMessage("[INFO] Message to topics: %s" %str(message.topics))
                data = json.dumps(message.toDict(), indent=4)
                
                if message.persist:
                    for t in message.topics:
                        self.__last_messages__[t] = data
                    
                    
                for user in self.__users__:
                    if len([t for t in message.topics if t in user.values.topics])>0:
                        await user.send(data)
        except Exception as e:
            LogMessage(f'[ERROR] Error on private message => {e}')
    
    async def queue(self, message:WebsocketMessage):
        try:
            if len(message.topics)>0:
                await self.send_private_message(message)
            else:
                await self.send_broadcast_message(message)
        except Exception as e:
            LogMessage(f'[ERROR] Error on send queue => {e}')

    async def serv(self, websocket, path):
        websocket.values = ArgsWebsocket.fromPath(path)
        await self.register(websocket)
        LogMessage(f"[INFO] {websocket.values.hashcode} start connection on topics: {websocket.values.topics}")
        while(1):
            try:
                await self.queue(WebsocketMessage.fromRecv(await websocket.recv()))
            except Exception as e:
                LogMessage(f"[ERROR] Error on websocket connection => {self.__class__.__name__} -> {str(e)} ")
                await self.unregister(websocket)
                break
    
    def start(self):
        LogMessage(f"[INFO] 🚀 Server Started: %s:%s" %(self.__uri__ or "0.0.0.0", str(self.__port__)))
        start_server = websockets.serve(self.serv, self.__uri__, self.__port__)
        asyncio.get_event_loop().run_until_complete(start_server)
        asyncio.get_event_loop().run_forever()
        
    
    def start_linux(self):
        LogMessage(f"[INFO] 🚀 Server Started: %s:%s" %(self.__uri__ or "127.0.0.1", str(self.__port__)))
        async def main():
            async with websockets.serve(self.serv, self.__uri__, self.__port__):
                await asyncio.Future()
        asyncio.run(main())


from datetime import datetime

class LogMessage:
    def __init__(self, message):
        self.__dt__ = datetime.now().strftime(" %d-%m-%Y %H:%M:%S -> ")
        self.__message__ = message
        print(self.getMessage())
        
    def getMessage(self):
        return f"{self.__dt__}{self.__message__}"
        
    def __str__(self) -> str:
        return self.getMessage()